# External Libraries
from unicodedata import east_asian_width
from wcwidth import wcwidth
import time, re, logging

from typing import Iterable
from prompt_toolkit import ANSI
from prompt_toolkit.application import get_app
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.formatted_text import to_formatted_text, fragment_list_to_text
from prompt_toolkit.formatted_text.base import OneStyleAndTextTuple
from prompt_toolkit.layout.processors import Processor, Transformation
from prompt_toolkit.application.current import get_app
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.document import Document
from prompt_toolkit.data_structures import Point
from prompt_toolkit.layout.controls import UIContent
from prompt_toolkit.lexers import Lexer
from prompt_toolkit.mouse_events import MouseButton, MouseEvent, MouseEventType
from prompt_toolkit.selection import SelectionType
from prompt_toolkit.buffer import Buffer, ValidationState

from prompt_toolkit.filters import (
    FilterOrBool,
)
from prompt_toolkit.formatted_text import (
    StyleAndTextTuples,
    to_formatted_text,
)
from prompt_toolkit.formatted_text.utils import fragment_list_to_text
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.key_binding.key_bindings import KeyBindingsBase
from prompt_toolkit.layout.containers import (
    Window,
    WindowAlign,
)
from prompt_toolkit.layout.controls import (
    BufferControl,
    FormattedTextControl,
)
from prompt_toolkit.layout.processors import (
    Processor,
)
from prompt_toolkit.lexers import Lexer
from prompt_toolkit.mouse_events import MouseEvent, MouseEventType
from prompt_toolkit.utils import get_cwidth
from prompt_toolkit.widgets import Button, MenuContainer, MenuItem
from prompt_toolkit.widgets.base import Border

from prompt_toolkit.layout.screen import _CHAR_CACHE, Screen, WritePosition
from prompt_toolkit.layout.utils import explode_text_fragments
from prompt_toolkit.formatted_text.utils import (
    fragment_list_to_text,
    fragment_list_width,
)

from .settings import Settings

class MudFormatProcessor(Processor):
    "在BufferControl中显示ANSI格式的处理器"

    def __init__(self) -> None:
        super().__init__()
        self.FULL_BLOCKS = set("▂▃▅▆▇▄█")
        self.SINGLE_LINES = set("┌└├┬┼┴╭╰─")
        self.DOUBLE_LINES = set("╔╚╠╦╪╩═")
        self.START_COLOR_REGX  = re.compile(r"^\[[\d;]+m")
        self.COLOR_REGX   = re.compile(r"\[[\d;]+m")
        self._color_start = ""
        self._color_correction = False

    def width_correction(self, line: str) -> str:
        new_str = []
        for ch in line:
            new_str.append(ch)
            if (east_asian_width(ch) in "FWA") and (wcwidth(ch) == 1):
                if ch in self.FULL_BLOCKS:
                    new_str.append(ch)
                elif ch in self.SINGLE_LINES:
                    new_str.append("─")
                elif ch in self.DOUBLE_LINES:
                    new_str.append("═")
                else:
                    new_str.append(' ')

        return "".join(new_str)
    
    def return_correction(self, line: str):
        return line.replace("\r", "").replace("\x00", "")
    
    def tab_correction(self, line: str):
        return line.replace("\t", " " * Settings.client["tabstop"])

    def color_correction(self, line: str):
        # 注：发现processer处理并非自上而下逐行处理的，因此不能使用这种颜色校正方式。
        if self._color_correction:
            other = self.COLOR_REGX.findall(line)

            line = f"{self._color_start}{line}"
            logging.debug(f"已校正增加颜色标志 {self._color_start}: {line}")

            if other:
                self._color_correction = False
                self._color_start = ""
                logging.debug(f"颜色校正结束: {line}")
        else:
            color = self.START_COLOR_REGX.findall(line)
            if color:
                other = self.COLOR_REGX.findall(line)
                if len(other) == 1:
                    self._color_correction = True
                    self._color_start = color[0]
                    logging.debug(f"获取到一个颜色开头 {color[0]}: {line}")

        return line

    def line_correction(self, line: str):
        # 处理\r符号（^M）
        line = self.return_correction(line)
        # 处理Tab(\r)符号（^I）
        line = self.tab_correction(line)
        # 处理颜色跨行问题。发现processer处理并非自上而下逐行处理的，因此不能使用这种颜色校正方式。
        # line = self.color_correction(line)
        # 美化（解决中文英文在Console中不对齐的问题）
        if Settings.client["beautify"]:
            line = self.width_correction(line)

        return line

    def apply_transformation(self, transformation_input):
        # 准备（先还原为str）
        line = fragment_list_to_text(transformation_input.fragments)
        line = self.line_correction(line)
        # 处理ANSI标记（生成FormmatedText）
        fragments = to_formatted_text(ANSI(line))
        return Transformation(fragments)

class SessionBuffer(Buffer):
    "继承自Buffer，为Session内容所修改，主要修改为只能在最后新增内容，并且支持分屏显示适配"

    def __init__(
        self,
    ):
        super().__init__()

        # 修改内容
        self.__text = ""
        self.__split = False
        
    def _set_text(self, value: str) -> bool:
        """set text at current working_index. Return whether it changed."""
        original_value = self.__text
        self.__text = value

        # Return True when this text has been changed.
        if len(value) != len(original_value):
            return True
        elif value != original_value:
            return True
        return False

    @property
    def text(self) -> str:
        return self.__text

    @text.setter
    def text(self, value: str) -> None:
        # SessionBuffer is only appendable

        if self.cursor_position > len(value):
            self.cursor_position = len(value)

        changed = self._set_text(value)

        if changed:
            self._text_changed()
            self.history_search_text = None

    @property
    def working_index(self) -> int:
        return 0

    @working_index.setter
    def working_index(self, value: int) -> None:
        pass

    def _text_changed(self) -> None:
        # Remove any validation errors and complete state.
        self.validation_error = None
        self.validation_state = ValidationState.UNKNOWN
        self.complete_state = None
        self.yank_nth_arg_state = None
        self.document_before_paste = None
        
        # 添加内容时，不取消选择
        #self.selection_state = None

        self.suggestion = None
        self.preferred_column = None

        # fire 'on_text_changed' event.
        self.on_text_changed.fire()

    def set_document(self, value: Document, bypass_readonly: bool = False) -> None:
        pass

    @property
    def split(self) -> bool:
        return self.__split
    
    @split.setter
    def split(self, value: bool) -> None:
        self.__split = value

    @property
    def is_returnable(self) -> bool:
        return False

    # End of <getters/setters>

    def save_to_undo_stack(self, clear_redo_stack: bool = True) -> None:
        pass

    def delete(self, count: int = 1) -> str:
        pass

    def insert_text(
        self,
        data: str,
        overwrite: bool = False,
        move_cursor: bool = True,
        fire_event: bool = True,
    ) -> None:
        # 始终在最后增加内容
        self.text += data
        
        # 分隔情况下，光标保持原位置不变，否则光标始终位于最后
        if not self.__split:
            # 若存在选择状态，则视情保留选择
            if self.selection_state:
                start = self.selection_state.original_cursor_position
                end = self.cursor_position
                row, col = self.document.translate_index_to_position(start)
                lastrow, col = self.document.translate_index_to_position(len(self.text))
                self.exit_selection()
                # 还没翻过半页的话，就重新选择上
                if lastrow - row < get_app().output.get_size().rows // 2 - 1:
                    self.cursor_position = len(self.text)
                    self.cursor_position = start
                    self.start_selection
                    self.cursor_position = end
                else:
                    self.cursor_position = len(self.text)
            else:
                self.cursor_position = len(self.text)
        else:
            pass
        

    def clear_half(self):
        "将Buffer前半段内容清除，并清除缓存"
        remain_lines = len(self.document.lines) // 2
        start = self.document.translate_row_col_to_index(remain_lines, 0)
        new_text = self.text[start:]

        del self.history
        self.history = InMemoryHistory()
        
        self.text = ""
        self._set_text(new_text)

        self._document_cache.clear()
        new_doc  = Document(text = new_text, cursor_position = len(new_text))
        self.reset(new_doc, False)
        self.__split = False

        return new_doc.line_count

    def undo(self) -> None:
        pass

    def redo(self) -> None:
        pass


class SessionBufferControl(BufferControl):
    def __init__(self, buffer: SessionBuffer = None, input_processors = None, include_default_input_processors: bool = True, lexer: Lexer = None, preview_search: FilterOrBool = False, focusable: FilterOrBool = True, search_buffer_control = None, menu_position = None, focus_on_click: FilterOrBool = False, key_bindings: KeyBindingsBase = None):
        # 将所属Buffer类型更改为SessionBuffer
        buffer = buffer or SessionBuffer()
        super().__init__(buffer, input_processors, include_default_input_processors, lexer, preview_search, focusable, search_buffer_control, menu_position, focus_on_click, key_bindings)
        self.buffer = buffer

    # def create_content(
    #     self, width: int, height: int, preview_search: bool = False
    # ) -> UIContent:
    #     """
    #     Create a UIContent.
    #     """
    #     buffer = self.buffer

    #     # Trigger history loading of the buffer. We do this during the
    #     # rendering of the UI here, because it needs to happen when an
    #     # `Application` with its event loop is running. During the rendering of
    #     # the buffer control is the earliest place we can achieve this, where
    #     # we're sure the right event loop is active, and don't require user
    #     # interaction (like in a key binding).
    #     buffer.load_history_if_not_yet_loaded()

    #     # Get the document to be shown. If we are currently searching (the
    #     # search buffer has focus, and the preview_search filter is enabled),
    #     # then use the search document, which has possibly a different
    #     # text/cursor position.)
    #     search_control = self.search_buffer_control
    #     preview_now = preview_search or bool(
    #         # Only if this feature is enabled.
    #         self.preview_search()
    #         and
    #         # And something was typed in the associated search field.
    #         search_control
    #         and search_control.buffer.text
    #         and
    #         # And we are searching in this control. (Many controls can point to
    #         # the same search field, like in Pyvim.)
    #         get_app().layout.search_target_buffer_control == self
    #     )

    #     if preview_now and search_control is not None:
    #         ss = self.search_state

    #         document = buffer.document_for_search(
    #             SearchState(
    #                 text=search_control.buffer.text,
    #                 direction=ss.direction,
    #                 ignore_case=ss.ignore_case,
    #             )
    #         )
    #     else:
    #         document = buffer.document

    #     get_processed_line = self._create_get_processed_line_func(
    #         document, width, height
    #     )
    #     self._last_get_processed_line = get_processed_line

    #     def translate_rowcol(row: int, col: int) -> Point:
    #         "Return the content column for this coordinate."
    #         return Point(x=get_processed_line(row).source_to_display(col), y=row)

    #     def get_line(i: int) -> StyleAndTextTuples:
    #         "Return the fragments for a given line number."
    #         fragments = get_processed_line(i).fragments

    #         # Add a space at the end, because that is a possible cursor
    #         # position. (When inserting after the input.) We should do this on
    #         # all the lines, not just the line containing the cursor. (Because
    #         # otherwise, line wrapping/scrolling could change when moving the
    #         # cursor around.)
    #         fragments = fragments + [("", " ")]
    #         return fragments

    #     content = UIContent(
    #         get_line=get_line,
    #         line_count=document.line_count,
    #         cursor_position=translate_rowcol(
    #             document.cursor_position_row, document.cursor_position_col
    #         ),
    #     )

    #     # If there is an auto completion going on, use that start point for a
    #     # pop-up menu position. (But only when this buffer has the focus --
    #     # there is only one place for a menu, determined by the focused buffer.)
    #     if get_app().layout.current_control == self:
    #         menu_position = self.menu_position() if self.menu_position else None
    #         if menu_position is not None:
    #             assert isinstance(menu_position, int)
    #             menu_row, menu_col = buffer.document.translate_index_to_position(
    #                 menu_position
    #             )
    #             content.menu_position = translate_rowcol(menu_row, menu_col)
    #         elif buffer.complete_state:
    #             # Position for completion menu.
    #             # Note: We use 'min', because the original cursor position could be
    #             #       behind the input string when the actual completion is for
    #             #       some reason shorter than the text we had before. (A completion
    #             #       can change and shorten the input.)
    #             menu_row, menu_col = buffer.document.translate_index_to_position(
    #                 min(
    #                     buffer.cursor_position,
    #                     buffer.complete_state.original_document.cursor_position,
    #                 )
    #             )
    #             content.menu_position = translate_rowcol(menu_row, menu_col)
    #         else:
    #             content.menu_position = None

    #     return content

    def mouse_handler(self, mouse_event: MouseEvent):
        """
        鼠标处理，修改内容包括：
        1. 在CommandLine获得焦点的时候，鼠标对本Control也可以操作
        2. 鼠标双击为选中行
        """
        buffer = self.buffer
        position = mouse_event.position

        # Focus buffer when clicked.
        cur_control = get_app().layout.current_control
        cur_buffer = get_app().layout.current_buffer
        # 这里时修改的内容
        if (cur_control == self) or (cur_buffer and cur_buffer.name == "input"):
            if self._last_get_processed_line:
                processed_line = self._last_get_processed_line(position.y)

                # Translate coordinates back to the cursor position of the
                # original input.
                xpos = processed_line.display_to_source(position.x)
                index = buffer.document.translate_row_col_to_index(position.y, xpos)

                # Set the cursor position.
                if mouse_event.event_type == MouseEventType.MOUSE_DOWN:
                    buffer.exit_selection()
                    buffer.cursor_position = index

                elif (
                    mouse_event.event_type == MouseEventType.MOUSE_MOVE
                    and mouse_event.button != MouseButton.NONE
                ):
                    # Click and drag to highlight a selection
                    if (
                        buffer.selection_state is None
                        and abs(buffer.cursor_position - index) > 0
                    ):
                        buffer.start_selection(selection_type=SelectionType.CHARACTERS)
                    buffer.cursor_position = index

                elif mouse_event.event_type == MouseEventType.MOUSE_UP:
                    # When the cursor was moved to another place, select the text.
                    # (The >1 is actually a small but acceptable workaround for
                    # selecting text in Vi navigation mode. In navigation mode,
                    # the cursor can never be after the text, so the cursor
                    # will be repositioned automatically.)
                    
                    if abs(buffer.cursor_position - index) > 1:
                        if buffer.selection_state is None:
                            buffer.start_selection(
                                selection_type=SelectionType.CHARACTERS
                            )
                        buffer.cursor_position = index

                    # Select word around cursor on double click.
                    # Two MOUSE_UP events in a short timespan are considered a double click.
                    double_click = (
                        self._last_click_timestamp
                        and time.time() - self._last_click_timestamp < 0.3
                    )
                    self._last_click_timestamp = time.time()

                    if double_click:
                        start = buffer.document.translate_row_col_to_index(position.y, 0)
                        end = buffer.document.translate_row_col_to_index(position.y + 1, 0) - 1
                        buffer.cursor_position = start
                        buffer.start_selection(selection_type=SelectionType.LINES)
                        buffer.cursor_position = end

                else:
                    # Don't handle scroll events here.
                    return NotImplemented

        # Not focused, but focusing on click events.
        else:
            if (
                self.focus_on_click()
                and mouse_event.event_type == MouseEventType.MOUSE_UP
            ):
                # Focus happens on mouseup. (If we did this on mousedown, the
                # up event will be received at the point where this widget is
                # focused and be handled anyway.)
                get_app().layout.current_control = self
            else:
                return NotImplemented

        return None

    def move_cursor_down(self) -> None:
        b = self.buffer
        b.cursor_position += b.document.get_cursor_down_position()

    def move_cursor_up(self) -> None:
        b = self.buffer
        b.cursor_position += b.document.get_cursor_up_position()

    def move_cursor_right(self, count = 1) -> None:
        b = self.buffer
        b.cursor_position += count

    def move_cursor_left(self, count = 1) -> None:
        b = self.buffer
        b.cursor_position -= count


class VSplitWindow(Window):
    "修改的分块窗口，向上翻页时，下半部保持最后数据不变"
 
    def _copy_body(
        self,
        ui_content: UIContent,
        new_screen: Screen,
        write_position: WritePosition,
        move_x: int,
        width: int,
        vertical_scroll: int = 0,
        horizontal_scroll: int = 0,
        wrap_lines: bool = False,
        highlight_lines: bool = False,
        vertical_scroll_2: int = 0,
        always_hide_cursor: bool = False,
        has_focus: bool = False,
        align: WindowAlign = WindowAlign.LEFT,
        get_line_prefix = None,
        isNotMargin = True,
    ):
        """
        Copy the UIContent into the output screen.
        Return (visible_line_to_row_col, rowcol_to_yx) tuple.

        :param get_line_prefix: None or a callable that takes a line number
            (int) and a wrap_count (int) and returns formatted text.
        """
        xpos = write_position.xpos + move_x
        ypos = write_position.ypos
        line_count = ui_content.line_count
        new_buffer = new_screen.data_buffer
        empty_char = _CHAR_CACHE["", ""]

        # Map visible line number to (row, col) of input.
        # 'col' will always be zero if line wrapping is off.
        visible_line_to_row_col: dict[int, tuple[int, int]] = {}

        # Maps (row, col) from the input to (y, x) screen coordinates.
        rowcol_to_yx: dict[tuple[int, int], tuple[int, int]] = {}

        def copy_line(
            line: StyleAndTextTuples,
            lineno: int,
            x: int,
            y: int,
            is_input: bool = False,
        ):
            """
            Copy over a single line to the output screen. This can wrap over
            multiple lines in the output. It will call the prefix (prompt)
            function before every line.
            """
            if is_input:
                current_rowcol_to_yx = rowcol_to_yx
            else:
                current_rowcol_to_yx = {}  # Throwaway dictionary.

            # Draw line prefix.
            if is_input and get_line_prefix:
                prompt = to_formatted_text(get_line_prefix(lineno, 0))
                x, y = copy_line(prompt, lineno, x, y, is_input=False)

            # Scroll horizontally.
            skipped = 0  # Characters skipped because of horizontal scrolling.
            if horizontal_scroll and is_input:
                h_scroll = horizontal_scroll
                line = explode_text_fragments(line)
                while h_scroll > 0 and line:
                    h_scroll -= get_cwidth(line[0][1])
                    skipped += 1
                    del line[:1]  # Remove first character.

                x -= h_scroll  # When scrolling over double width character,
                # this can end up being negative.

            # Align this line. (Note that this doesn't work well when we use
            # get_line_prefix and that function returns variable width prefixes.)
            if align == WindowAlign.CENTER:
                line_width = fragment_list_width(line)
                if line_width < width:
                    x += (width - line_width) // 2
            elif align == WindowAlign.RIGHT:
                line_width = fragment_list_width(line)
                if line_width < width:
                    x += width - line_width

            col = 0
            wrap_count = 0
            for style, text, *_ in line:
                new_buffer_row = new_buffer[y + ypos]

                # Remember raw VT escape sequences. (E.g. FinalTerm's
                # escape sequences.)
                if "[ZeroWidthEscape]" in style:
                    new_screen.zero_width_escapes[y + ypos][x + xpos] += text
                    continue

                for c in text:
                    char = _CHAR_CACHE[c, style]
                    char_width = char.width

                    # Wrap when the line width is exceeded.
                    if wrap_lines and x + char_width > width:
                        visible_line_to_row_col[y + 1] = (
                            lineno,
                            visible_line_to_row_col[y][1] + x,
                        )
                        y += 1
                        wrap_count += 1
                        x = 0

                        # Insert line prefix (continuation prompt).
                        if is_input and get_line_prefix:
                            prompt = to_formatted_text(
                                get_line_prefix(lineno, wrap_count)
                            )
                            x, y = copy_line(prompt, lineno, x, y, is_input=False)

                        new_buffer_row = new_buffer[y + ypos]

                        if y >= write_position.height:
                            return x, y  # Break out of all for loops.

                    # Set character in screen and shift 'x'.
                    if x >= 0 and y >= 0 and x < width:
                        new_buffer_row[x + xpos] = char

                        # When we print a multi width character, make sure
                        # to erase the neighbours positions in the screen.
                        # (The empty string if different from everything,
                        # so next redraw this cell will repaint anyway.)
                        if char_width > 1:
                            for i in range(1, char_width):
                                new_buffer_row[x + xpos + i] = empty_char

                        # If this is a zero width characters, then it's
                        # probably part of a decomposed unicode character.
                        # See: https://en.wikipedia.org/wiki/Unicode_equivalence
                        # Merge it in the previous cell.
                        elif char_width == 0:
                            # Handle all character widths. If the previous
                            # character is a multiwidth character, then
                            # merge it two positions back.
                            for pw in [2, 1]:  # Previous character width.
                                if (
                                    x - pw >= 0
                                    and new_buffer_row[x + xpos - pw].width == pw
                                ):
                                    prev_char = new_buffer_row[x + xpos - pw]
                                    char2 = _CHAR_CACHE[
                                        prev_char.char + c, prev_char.style
                                    ]
                                    new_buffer_row[x + xpos - pw] = char2

                        # Keep track of write position for each character.
                        current_rowcol_to_yx[lineno, col + skipped] = (
                            y + ypos,
                            x + xpos,
                        )

                    col += 1
                    x += char_width
            return x, y

        # Copy content.
        def copy() -> int:
            y = -vertical_scroll_2        
            lineno = vertical_scroll
            
            total = write_position.height
            upper = (total - 1) // 2
            below = total - upper - 1
            
            if lineno + total < line_count:
                if isinstance(self.content, SessionBufferControl):
                    b = self.content.buffer
                    b.split = True

                while y < upper and lineno < line_count:
                    line = ui_content.get_line(lineno)
                    visible_line_to_row_col[y] = (lineno, horizontal_scroll)
                    x = 0
                    x, y = copy_line(line, lineno, x, y, is_input=True)
                    lineno += 1
                    y += 1

                x = 0
                x, y = copy_line([("","-"*width)], lineno, x, y, is_input=False)
                y += 1

                lineno = line_count - below
                while y < total and lineno < line_count:
                    line = ui_content.get_line(lineno)
                    visible_line_to_row_col[y] = (lineno, horizontal_scroll)
                    x = 0
                    x, y = copy_line(line, lineno, x, y, is_input=True)
                    lineno += 1
                    y += 1

                return y
                
            else:
                if isNotMargin and isinstance(self.content, SessionBufferControl):
                    b = self.content.buffer
                    b.split = False

                while y < write_position.height and lineno < line_count:
                    # Take the next line and copy it in the real screen.
                    line = ui_content.get_line(lineno)

                    visible_line_to_row_col[y] = (lineno, horizontal_scroll)

                    # Copy margin and actual line.
                    x = 0
                    x, y = copy_line(line, lineno, x, y, is_input=True)

                    lineno += 1
                    y += 1
                return y
        
            

        copy()

        def cursor_pos_to_screen_pos(row: int, col: int) -> Point:
            "Translate row/col from UIContent to real Screen coordinates."
            try:
                y, x = rowcol_to_yx[row, col]
            except KeyError:
                # Normally this should never happen. (It is a bug, if it happens.)
                # But to be sure, return (0, 0)
                return Point(x=0, y=0)

                # raise ValueError(
                #     'Invalid position. row=%r col=%r, vertical_scroll=%r, '
                #     'horizontal_scroll=%r, height=%r' %
                #     (row, col, vertical_scroll, horizontal_scroll, write_position.height))
            else:
                return Point(x=x, y=y)

        # Set cursor and menu positions.
        if ui_content.cursor_position:
            screen_cursor_position = cursor_pos_to_screen_pos(
                ui_content.cursor_position.y, ui_content.cursor_position.x
            )

            if has_focus:
                new_screen.set_cursor_position(self, screen_cursor_position)

                if always_hide_cursor:
                    new_screen.show_cursor = False
                else:
                    new_screen.show_cursor = ui_content.show_cursor

                self._highlight_digraph(new_screen)

            if highlight_lines:
                self._highlight_cursorlines(
                    new_screen,
                    screen_cursor_position,
                    xpos,
                    ypos,
                    width,
                    write_position.height,
                )

        # Draw input characters from the input processor queue.
        if has_focus and ui_content.cursor_position:
            self._show_key_processor_key_buffer(new_screen)

        # Set menu position.
        if ui_content.menu_position:
            new_screen.set_menu_position(
                self,
                cursor_pos_to_screen_pos(
                    ui_content.menu_position.y, ui_content.menu_position.x
                ),
            )

        # Update output screen height.
        new_screen.height = max(new_screen.height, ypos + write_position.height)

        return visible_line_to_row_col, rowcol_to_yx

    def _copy_margin(
        self,
        margin_content: UIContent,
        new_screen: Screen,
        write_position: WritePosition,
        move_x: int,
        width: int,
    ) -> None:
        """
        Copy characters from the margin screen to the real screen.
        """
        xpos = write_position.xpos + move_x
        ypos = write_position.ypos

        margin_write_position = WritePosition(xpos, ypos, width, write_position.height)
        self._copy_body(margin_content, new_screen, margin_write_position, 0, width, isNotMargin=False)

    def _scroll_down(self) -> None:
        "向下滚屏，处理屏幕分隔"
        info = self.render_info

        if info is None:
            return

        if isinstance(self.content, SessionBufferControl):
            b = self.content.buffer
            d = b.document

            b.exit_selection()
            cur_line = d.cursor_position_row
            
            # # 向下滚动时，如果存在自动折行情况，要判断本行被折成了几行，在行内时要逐行移动（此处未调试好）
            # cur_col  = d.cursor_position_col
            # line = d.current_line
            # line_width = len(line)
            # line_start = d.translate_row_col_to_index(cur_line, 0)
            # screen_width = info.window_width

            # offset_y = cur_col // screen_width
            # wraplines = math.ceil(1.0 * line_width / screen_width)

            if cur_line < info.content_height:
                
                # if offset_y < wraplines:                                    # add
                #     self.content.move_cursor_right(screen_width)            # add
                # else:                                                       # add

                self.content.move_cursor_down()
                self.vertical_scroll = cur_line + 1

            firstline = d.line_count - len(info.displayed_lines)
            if cur_line >= firstline:
                b.cursor_position = len(b.text)

    def _scroll_up(self) -> None:
        "向上滚屏，处理屏幕分隔"
        info = self.render_info

        if info is None:
            return

        #if info.cursor_position.y >= 1:
        if isinstance(self.content, SessionBufferControl):
            b = self.content.buffer
            d = b.document

            b.exit_selection()
            cur_line = d.cursor_position_row
            if cur_line > d.line_count - len(info.displayed_lines):
                firstline = d.line_count - len(info.displayed_lines)
                newpos = d.translate_row_col_to_index(firstline, 0)
                b.cursor_position = newpos
                cur_line = d.cursor_position_row
                self.vertical_scroll = cur_line

            elif cur_line > 0:
                self.content.move_cursor_up()
                self.vertical_scroll = cur_line - 1


class EasternButton(Button):
    "解决增加中文等东亚全宽字符后不对齐问题"

    def _get_text_fragments(self) -> StyleAndTextTuples:
        # 主要改动在这里
        width = self.width - (
            get_cwidth(self.left_symbol) + get_cwidth(self.right_symbol)
        ) - (get_cwidth(self.text) - len(self.text))


        text = (f"{{:^{width}}}").format(self.text)

        def handler(mouse_event: MouseEvent) -> None:
            if (
                self.handler is not None
                and mouse_event.event_type == MouseEventType.MOUSE_UP
            ):
                self.handler()

        return [
            ("class:button.arrow", self.left_symbol, handler),
            #("[SetCursorPosition]", ""),
            ("class:button.text", text, handler),
            ("class:button.arrow", self.right_symbol, handler),
        ]

class EasternMenuContainer(MenuContainer):
    "解决增加中文等东亚全宽字符后不对齐问题"

    def _submenu(self, level: int = 0) -> Window:
        def get_text_fragments() -> StyleAndTextTuples:
            result: StyleAndTextTuples = []
            if level < len(self.selected_menu):
                menu = self._get_menu(level)
                if menu.children:
                    result.append(("class:menu", Border.TOP_LEFT))
                    result.append(("class:menu", Border.HORIZONTAL * (menu.width + 4)))
                    result.append(("class:menu", Border.TOP_RIGHT))
                    result.append(("", "\n"))
                    try:
                        selected_item = self.selected_menu[level + 1]
                    except IndexError:
                        selected_item = -1

                    def one_item(
                        i: int, item: MenuItem
                    ) -> Iterable[OneStyleAndTextTuple]:
                        def mouse_handler(mouse_event: MouseEvent) -> None:
                            if item.disabled:
                                # The arrow keys can't interact with menu items that are disabled.
                                # The mouse shouldn't be able to either.
                                return
                            hover = mouse_event.event_type == MouseEventType.MOUSE_MOVE
                            if (
                                mouse_event.event_type == MouseEventType.MOUSE_UP
                                or hover
                            ):
                                app = get_app()
                                if not hover and item.handler:
                                    app.layout.focus_last()
                                    item.handler()
                                else:
                                    self.selected_menu = self.selected_menu[
                                        : level + 1
                                    ] + [i]

                        if i == selected_item:
                            yield ("[SetCursorPosition]", "")
                            style = "class:menu-bar.selected-item"
                        else:
                            style = ""

                        yield ("class:menu", Border.VERTICAL)
                        if item.text == "-":
                            yield (
                                style + "class:menu-border",
                                f"{Border.HORIZONTAL * (menu.width + 3)}",
                                mouse_handler,
                            )
                        else:
                            # 主要改动在这里，其他地方都未更改.
                            adj_width = menu.width + 3 - (get_cwidth(item.text) - len(item.text))
                            yield (
                                style,
                                f" {item.text}".ljust(adj_width),
                                mouse_handler,
                            )

                        if item.children:
                            yield (style, ">", mouse_handler)
                        else:
                            yield (style, " ", mouse_handler)

                        if i == selected_item:
                            yield ("[SetMenuPosition]", "")
                        yield ("class:menu", Border.VERTICAL)

                        yield ("", "\n")

                    for i, item in enumerate(menu.children):
                        result.extend(one_item(i, item))

                    result.append(("class:menu", Border.BOTTOM_LEFT))
                    result.append(("class:menu", Border.HORIZONTAL * (menu.width + 4)))
                    result.append(("class:menu", Border.BOTTOM_RIGHT))
            return result

        return Window(FormattedTextControl(get_text_fragments), style="class:menu")



class MenuItem:
    def __init__(
        self,
        text: str = "",
        handler = None,
        children = None,
        shortcut = None,
        disabled: bool = False,
    ) -> None:
        self.text = text
        self.handler = handler
        self.children = children or []
        self.shortcut = shortcut
        self.disabled = disabled
        self.selected_item = 0

    @property
    def width(self) -> int:
        if self.children:
            return max(get_cwidth(c.text) for c in self.children)
        else:
            return 0


class DotDict(dict):
    """
    可以通过点.访问内部key-value对的字典。此类型继承自dict。

    - 由于继承关系，此类型可以使用所有dict可以使用的方法
    - 额外增加的点.访问方法使用示例如下

    示例:
        .. code:: Python

            mydict = DotDict()
            
            # 以下写内容访问等价
            mydict["key1"] = "value1"
            mydict.key1 = "value1"

            # 以下读访问等价
            val = mydict["key1"]
            val = mydict.key1
    """

    def __getattr__(self, __key):
        if (not __key in self.__dict__) and (not __key.startswith("__")):
            return self.__getitem__(__key)

    def __setattr__(self, __name: str, __value):
        if __name in self.__dict__:
            object.__setattr__(self, __name, __value)
        else:
            self.__setitem__(__name, __value)

    def __getstate__(self):
        return self
    
    def __setstate__(self, state):
        self.update(state)


        


# 示例脚本：如何在PyMud中玩PKUXKX

import webbrowser
from pymud import Alias, Trigger, SimpleCommand, Timer, SimpleTrigger, SimpleAlias

# 在PyMud中，使用#load {filename}可以加载对应的配置作为脚本文件以提供支撑。支持多脚本加载
# 本示例脚本对PyMud支持的变量(Variable)、触发器(Trigger，包含单行与多行触发)、别名(Alias)、定时器(Timer)、命令(Command，本示例中使用了SimpleCommand子类)都进行了代码示例
# 使用#load {filename}加载的配置文件中，若有一个类型名为Coniguration，则在#load操作时，会自动创建此类型；若没有Configuration类，则仅将文件引入
# 例如，加载本文件指定的配置，则使用 #load pymud.pkuxkx即可

# PyMud中，触发器Trigger、别名Alias、命令Command，都是匹配对象(MatchObject)的子类，使用同一种处理逻辑
# 匹配对象，意味着有匹配的pattern。在匹配对象成功后，会调用对象的onSuccess方法
# Trigger、Alias仅有成功，即仅onSuccess方法会被调用，该方法参数参考了MushClient，会传递name, line, wildcards三个参数，含义与MushClient相同；


class Configuration:

    # hpbrief long情况下的含义
    HP_KEYS = (
        "combat_exp", "potential", "max_neili", "neili", "max_jingli", "jingli", 
        "max_qi", "eff_qi", "qi", "max_jing", "eff_jing", "jing", 
        "vigour/qi", "vigour/yuan", "food", "water", "fighting", "busy"
        )

    # 类的构造函数，传递参数session，是会话本身
    def __init__(self, session) -> None:
        self.session = session
        self._triggers = {}
        self._commands = {}
        self._aliases  = {}
        self._timers   = {}

        self._initTriggers()
        self._initCommands()
        self._initAliases()
        self._initTimers()

    def _initTriggers(self):
        '''
        初始化触发器。
        本示例中创建了2个触发器，分别对应fullme的链接触发自动打开浏览器访问该网址，以及hpbrief命令的触发器
        '''
        # Trigger的构造函数中，给定的位置参数仅有session(会话)和pattern(匹配模式)两个，其他所有参数都是使用命名参数进行实现
        # 支持的命名参数详细可参见 BaseObject与MatchObject 的构造方法，此处简单列举
        # id         : 唯一标识，不指定时自动生成
        # group      : 组名，不指定时为空
        # enabled    : 使能状态，默认True
        # priority   : 优先级，默认100，越小越高
        # oneShot    : 单次匹配，默认False
        # ignoreCase : 忽略大小写，默认False
        # isRegExp   : 正则表达式模式，默认True
        # keepEval   : 持续匹配，默认False
        # raw        : 原始匹配方式，默认False。原始匹配方式下，不对VT100下的ANSI颜色进行解码，因此可以匹配颜色；正常匹配仅匹配文本
        
        # 1. fullme的链接对应的触发器，匹配URL
        # 当匹配成功后，调用ontri_webpage
        self._triggers["tri_webpage"] = self.tri_webpage = Trigger(self.session, id = 'tri_webpage', patterns = r'^http://fullme.pkuxkx.net/robot.php.+$', group = "sys", onSuccess =  lambda id, line, wildcards: webbrowser.open(line))
        # 2. fullme的链接对应的触发器，因为要进行多行匹配（3行），因此匹配模式pattern为3个正则表达式模式构成的元组（所有列表类型均可识别），无需像MushClient一样要指定multiline标识和linesToMatch数量
        # 当匹配成功后，调用ontri_hpbrief
        # 特别说明：此处的hpbrief触发匹配，需要set hpbrief long后才可以支持
        self._triggers["tri_hp"]      = self.tri_hp      = Trigger(self.session, id = 'tri_hpbrief', patterns = (r'^[> ]*#(\d+.?\d*[KM]?),(\d+),(\d+),(\d+),(\d+),(\d+)$', r'^[> ]*#(\d+),(\d+),(\d+),(\d+),(\d+),(\d+)$', r'^[> ]*#(\d+),(\d+),(-?\d+),(-?\d+),(\d+),(\d+)$',), group = "sys", onSuccess = self.ontri_hpbrief)
        
        # 3. 现在支持简单Trigger了，例如
        self._triggers["tri_gem"] = SimpleTrigger(self.session ,r'^[> ]*从.+身上.+[◎☆★].+', "pack gem", group = "sys")

        self.session.addTriggers(self._triggers)


    def _initCommands(self):
        '''初始化命令，本示例中创建了1个命令，支持hpbrief命令'''

        # Command是异步执行的命令，可以理解为Alias+Trigger+Timer的组合。在MUD中发出一条命令后，会有成功、失败、超时的不同状态，在这三种状态下，会分别调用onSuccess、onFailure、onTimeout方法
        # 举个Command的应用例子说明：加入把移动（s/e/n/w等等）实现为一个Command。
        # 1. 当向某个方向移动时，成功的时候会移动到下一个房间；
        # 2. 不成功的时候，会出现“这个方向没有出路”等描述；
        # 3. 而当角色处于晕倒状态时，移动命令是不会有任何反应的，超出设定超时时间之后，会调用onTimeout
        # 在上述实现情况下，我们在执行命令时，可以明确的根据命令的执行结果再判断下一步该做什么
        # 本示例使用了简单命令SimpleCommand，其在MatchObject的基础上，增加了以下参数：
        # 1. succ_tri: 命令执行成功时的触发器，不能为空
        # 2. fail_tri: 命令执行失败时的触发器，可以为空
        # 3. retry_tri: 需要重新尝试命令的触发器，可以为空（仍以移动为例，当向某个方向移动，出现“你现在正忙着呢”之后，可以在等待2s之后，再次尝试该命令，知道到达最大尝试次数

        # 命令可以同步调用，也可以在异步函数（async）中使用await语法异步调用
        # 例如，下面的hpbrief可以在这样使用：
        # self.session.exec_command("hpbrief")
        # self.session.exec_command_after(2, "hpbrief")
        # await self.cmd_hpbrief.execute("hpbrief")

        # 异步实现意味着，在函数实现过程中可以以循环实现，而不是以回调实现，有利于代码的可读性
        # 假设已经实现了一个 cmd_move 的 Command，现在要从ct 执行"s;s;w"行走指令到达春来茶馆，然后根据当前的hpbrief结果判断是否需要drink，然后走回中央广场，可以在函数中这样实现：
        # async def gotodrink(self):
        #     for step in "s;s;w".split(";"):
        #         await self.cmd_move.execute(step)
        #         await self.cmd_hpbrief.execute("hpbrief")
        #     await asyncio.sleep(1)
        #     water = self.session.getVariable("water")
        #     if int(water) < 300:
        #         self.session.writeline("drink")
        #     await asyncio.sleep(1)
        #     for step in "e;n;n".split(";"):
        #         await self.cmd_move.execute(step)


        self._commands['cmd_hpbrief']    = self.cmd_hpbrief     = SimpleCommand(self.session, id = "cmd_hpbrief", patterns = "^hpbrief$", succ_tri = self.tri_hp, group = "status", onSuccess = self.oncmd_hpbrief)
        self.session.addCommands(self._commands)

    def _initAliases(self):
        '''初始化别名，本示例中创建了1个别名，是get xxx from corpse'''

        #  get xxx from corpse的别名操作，匹配成功后会自动调用getfromcorpse函数
        #  例如， gp silver 相当于 get silver from corpse
        self._aliases['ali_get'] = Alias(self.session, r"^gp\s(.+)$", id = "ali_get", onSuccess = self.getfromcorpse)

        # 3. 现在支持简单Alias了，在其中也可以支持#wait（缩写为#wa操作）等待，当然，Trigger也支持
        # 从扬州中心广场到西门的行走，每步中间插入100ms等待
        self._aliases["ali_yz_xm"] = SimpleAlias(self.session ,'^yz_xm$', "w;#wa 100;w;#wa 100;w;#wa 100;w", group = "sys")

        self.session.addAliases(self._aliases)

    def _initTimers(self):
        '''初始化定时器，本示例中创建了1个定时器，每隔2秒打印信息'''

        self._timers["tm_test"] = self.tm_test = Timer(self.session, timeout = 2, id = "tm_test", onSuccess = self.onTimer)
        self.session.addTimers(self._timers)

    def getfromcorpse(self, name, line, wildcards):
        cmd = f"get {wildcards[0]} from corpse"
        self.session.writeline(cmd)

    def onTimer(self, name, *args, **kwargs):
        self.session.info("每2秒都会打印本信息", "定时器测试")

    def ontri_hpbrief(self, name, line, wildcards):
        self.session.setVariables(self.HP_KEYS, wildcards)

    def oncmd_hpbrief(self, name, cmd, line, wildcards):
        # 为了节省服务器资源，应使用hpbrief来代替hp指令
        # 但是hpbrief指令的数据看起来太麻烦，所以将hpbrief的一串数字输出成类似hp的样式
        # ┌───个人状态────────────────────┬─────────────────────────────┐
        # │【精神】 1502    / 1502     [100%]    │【精力】 4002    / 4002    (+   0)    │
        # │【气血】 2500    / 2500     [100%]    │【内力】 5324    / 5458    (+   0)    │
        # │【真气】 0       / 0        [  0%]    │【禅定】 101%               [正常]    │
        # │【食物】 222     / 400      [缺食]    │【潜能】 36,955                       │
        # │【饮水】 247     / 400      [缺水]    │【经验】 2,341,005                    │
        # ├─────────────────────────────┴─────────────────────────────┤
        # │【状态】 健康、怒                                                             │
        # └────────────────────────────────────────────北大侠客行────────┘
        var1 = self.session.getVariables(("jing", "effjing", "maxjing", "jingli", "maxjingli"))
        line1 = "【精神】 {0:<8} [{5:3.0f}%] / {1:<8} [{2:3.0f}%]  |【精力】 {3:<8} / {4:<8} [{6:3.0f}%]".format(var1[0], var1[1], 100 * float(var1[1]) / float(var1[2]), var1[3], var1[4], 100 * float(var1[0]) / float(var1[2]), 100 * float(var1[3]) / float(var1[4]))
        var2 = self.session.getVariables(("qi", "effqi", "maxqi", "neili", "maxneili"))
        line2 = "【气血】 {0:<8} [{5:3.0f}%] / {1:<8} [{2:3.0f}%]  |【内力】 {3:<8} / {4:<8} [{6:3.0f}%]".format(var2[0], var2[1], 100 * float(var2[1]) / float(var2[2]), var2[3], var2[4], 100 * float(var2[0]) / float(var2[2]), 100 * float(var2[3]) / float(var2[4]))
        var3 = self.session.getVariables(("food", "water", "exp", "pot", "fighting", "busy"))
        line3 = "【食物】 {0:<4} 【饮水】{1:<4} 【经验】{2:<9} 【潜能】{3:<10}【{4}】【{5}】".format(var3[0], var3[1], var3[2], var3[3],  "未战斗" if var3[4] == "0" else "战斗中", "不忙" if var3[5] == "0" else "忙")
        self.session.info(line1, "状态")
        self.session.info(line2, "状态")
        self.session.info(line3, "状态")



import os, sys, json, platform, shutil, logging, argparse
from pathlib import Path
from .pymud import PyMudApp
from .settings import Settings

CFG_TEMPLATE = {
    "client": {
        "buffer_lines"      : 5000,                 # 保留缓冲行数

        "interval"          : 10,                   # 在自动执行中，两次命令输入中的间隔时间（ms）
        "auto_connect"      : True,                 # 创建会话后，是否自动连接
        "auto_reconnect"    : False,                # 在会话异常断开之后，是否自动重连
        "var_autosave"      : True,                 # 断开时自动保存会话变量
        "var_autoload"      : True,                 # 初始化时自动加载会话变量

        "echo_input"        : False,
        "beautify"          : True,                 # 专门为解决控制台下PKUXKX字符画对不齐的问题

        "status_display"    : 1,                    # 状态窗口显示情况设置，0-不显示，1-显示在下方，2-显示在右侧
        "status_height"     : 4,                    # 下侧状态栏的高度
        "status_width"      : 30,                   # 右侧状态栏的宽度

    },
    "sessions" : {
        "pkuxkx" : {
            "host" : "mud.pkuxkx.net",
            "port" : "8081",
            "encoding" : "utf8",
            "autologin" : "{0};{1}",
            "default_script": "examples",
            "chars" : {
                "display_title" : ["yourid", "yourpassword", ""],
            }
        }
    },
    "keys" : {
        "f3"    : "#ig",
        "f4"    : "#clear",
        "f11"   : "#close",
        "f12"   : "#exit",
    }
}

def init_pymud_env(args):
    print(f"欢迎使用PyMUD, 版本{Settings.__version__}. 使用PyMUD时, 建议建立一个新目录（任意位置），并将自己的脚本以及配置文件放到该目录下.")
    print("即将开始为首次运行初始化环境...")
    
    dir = args.dir
    if dir:
        print(f"你已经指定了创建脚本的目录为 {args.dir}")
        dir = Path(dir)
    else:
        dir = Path.home().joinpath('pkuxkx')

        system = platform.system().lower()
        dir_enter = input(f"检测到当前系统为 {system}, 请指定游戏脚本的目录（若目录不存在会自动创建），直接回车表示使用默认值 [{dir}]:")
        if dir_enter:
            dir = Path(dir_enter)
                      
    if dir.exists() and dir.is_dir():
        print(f'检测到给定目录 {dir} 已存在，切换至此目录...')
    else:
        print(f'检测到给定目录 {dir} 不存在，正在创建并切换至目录...')
        dir.mkdir()
    
    os.chdir(dir)

    if os.path.exists('pymud.cfg'):
        print(f'检测到脚本目录下已存在pymud.cfg文件，将直接使用此文件进入PyMUD...')
    else:
        print(f'检测到脚本目录下不存在pymud.cfg文件，将使用默认内容创建该配置文件...')
        with open('pymud.cfg', mode = 'x') as fp:
            fp.writelines(json.dumps(CFG_TEMPLATE, indent = 4))

    if not os.path.exists('examples.py'):
        from pymud import pkuxkx
        module_dir = pkuxkx.__file__
        shutil.copyfile(module_dir, 'examples.py')
        print(f'已将样例脚本拷贝至脚本目录，并加入默认配置文件')

    print(f"后续可自行修改 {dir} 目录下的 pymud.cfg 文件以进行配置。")
    if system == "windows":
        print(f"后续运行PyMUD， 请在 {dir} 目录下键入命令： python -m pymud")
    else:
        print(f"后续运行PyMUD， 请在 {dir} 目录下键入命令： python3 -m pymud")

    input('所有内容已初始化完毕, 请按回车进入PyMUD.')

    startApp(args)

def startApp(args):
    startup_path = Path(args.startup_dir).resolve()
    sys.path.append(f"{startup_path}")
    os.chdir(startup_path)

    if args.debug:
        logging.basicConfig(level = logging.NOTSET,
            format = '%(asctime)s %(name)-12s %(levelname)-8s %(message)s',
            datefmt = '%m-%d %H:%M',
            filename = args.logfile,
            filemode = 'a' if args.filemode else 'w',
            encoding = "utf-8"
            )
        
    else:
        logging.basicConfig(level = logging.NOTSET,
            format = '%(asctime)s %(name)-12s: %(message)s',
            datefmt = '%m-%d %H:%M',
            handlers = [logging.NullHandler()],
            )

    cfg = startup_path.joinpath("pymud.cfg")
    cfg_data = None
    if os.path.exists(cfg):
        with open(cfg, "r", encoding="utf8", errors="ignore") as fp:
            cfg_data = json.load(fp)

    app = PyMudApp(cfg_data)
    app.run()

def main():
    parser = argparse.ArgumentParser(prog = "pymud", description = "PyMUD命令行参数帮助")
    subparsers = parser.add_subparsers(help = 'init用于初始化运行环境')

    par_init = subparsers.add_parser('init', description = '初始化pymud运行环境, 包括建立脚本目录, 创建默认配置文件, 创建样例脚本等.')
    par_init.add_argument('-d', '--dir', dest = 'dir', metavar = 'dir', type = str, default = '', help = '指定构建脚本目录的名称, 不指定时会根据操作系统选择不同默认值')
    par_init.set_defaults(func = init_pymud_env)

    parser.add_argument('-d', '--debug', dest = 'debug', action = 'store_true', default = False, help = '指定以调试模式进入PyMUD。此时，系统log等级将设置为logging.NOTSET, 所有log数据均会被记录。默认不启用。')
    parser.add_argument('-l', '--logfile', dest = 'logfile', metavar = 'logfile', default = 'pymud.log', help = '指定调试模式下记录文件名，不指定时，默认为当前目录下的pymud.log')
    parser.add_argument('-a', '--appendmode', dest = 'filemode', action = 'store_true', default = True, help = '指定log文件的访问模式是否为append尾部添加模式，默认为True。当为False时，使用w模式，即每次运行清空之前记录')
    parser.add_argument('-s', '--startup_dir', dest = 'startup_dir', metavar = 'startup_dir', default = '.', help = '指定启动目录，默认为当前目录。使用该参数可以在任何目录下，通过指定脚本目录来启动')

    args=parser.parse_args()

    if hasattr(args, 'func'):
        args.func(args)
    else:
        startApp(args)

if __name__ == "__main__":
    main()